from neumai.Shared.NeumSearch import NeumSearchResult
from neumai.Shared.NeumVector import NeumVector
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from abc import ABC, abstractmethod
from typing import List

# Eventually this is an abstract class where specific implementations of it implement from it.
# Like, PineconeConnector which extends/implements from SinkConnector
class SinkConnector(ABC):
    def __init__(self, sink_information: dict = {}):
        self.sink_information = sink_information
    
    @property
    @abstractmethod
    def sink_name(self) -> str:
        pass
    
    @property
    @abstractmethod
    def requiredProperties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optionalProperties(self) -> List[str]:
        pass

    @abstractmethod
    def validation(self) -> bool:
        """Validate sink setup"""

    @abstractmethod
    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        """Store vectors with a given service"""

    @abstractmethod
    def search(self, vector:List[float], number_of_results:int, pipeline_id:str) -> List[NeumSearchResult]:
        """Search vectors for a given service"""
    
    @abstractmethod
    def info(self, pipeline_id:str) -> NeumSinkInfo:
        """Get information about what is stores in the sink"""

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['sink_name'] = self.sink_name
        json_to_return['sink_information'] = self.sink_information
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['sink_name'] = self.sink_name
        json_to_return['sink_information'] = self.sink_information
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['requiredProperties'] = self.requiredProperties
        json_to_return['optionalProperties'] = self.optionalProperties
        return json_to_return


from neumai.DataConnectors.DataConnector import DataConnector
from neumai.DataConnectors import (
    S3Connector,
    NeumWebsiteConnector,
    AzureBlobConnector,
    SupabaseStorageConnector,
    PostgresConnector,
    NeumFileConnector
)
from neumai.Shared.Selector import Selector
from starlette.exceptions import HTTPException

# Helper function to map connector configuration to object type

def as_connector(dct:dict):
    if dct == None:
        raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
    connector_name = dct.get("connector_name", None)
    connector_information = dct.get("connector_information", None)
    selector = Selector.as_selector(dct.get("selector", None))
    if connector_name == "AzureBlobConnector":
        return AzureBlobConnector(connector_information=connector_information, selector=selector)
    elif connector_name == "NeumWebsiteConnector":
        return NeumWebsiteConnector(connector_information=connector_information, selector=selector)
    elif connector_name == "S3Connector":
        return S3Connector(connector_information=connector_information, selector=selector)
    elif connector_name == "SupabaseStorageConnector":
        return SupabaseStorageConnector(connector_information=connector_information, selector=selector)
    elif connector_name == "PostgresConnector":
        return PostgresConnector(connector_information=connector_information, selector=selector)
    elif connector_name == "NeumFileConnector":
        return NeumFileConnector(connector_information=connector_information, selector=selector)
    return None


from .PipelineRun import PipelineRun
from .TriggerSchedule import TriggerSchedule
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from neumai.Sources.SourceConnector import SourceConnector
from neumai.EmbedConnectors.EmbedHelper import as_embed
from neumai.SinkConnectors.SinkHelper import as_sink
from neumai.Shared.NeumVector import NeumVector
from neumai.Shared.NeumSearch import NeumSearchResult
from typing import List
import uuid

accepted_trigger_sync_types: List = []
class Pipeline(object):
    def __init__(self, 
                sources:  List[SourceConnector], # Change to only support a list of Sources for V2
                embed: EmbedConnector, 
                sink: SinkConnector, 
                name:str = None,
                id: str = str(uuid.uuid4()), 
                created: float = None, 
                updated: float = None,
                trigger_schedule: TriggerSchedule = None, 
                latest_run: PipelineRun = None, 
                owner: str = None, 
                is_deleted: bool = False):
        self.id = id
        self.name = name
        self.created = created
        self.updated = updated
        self.sources = sources
        self.embed = embed
        self.sink = sink
        self.trigger_schedule = trigger_schedule
        self.latest_run = latest_run 
        self.owner = owner
        self.is_deleted = is_deleted

    def validate(self) -> bool:
        """Running validation for each connector"""
        try:
            for source in self.sources:
                source.validation()
            self.embed.validation()
            self.sink.validation()
            return True
        except Exception as e:
            raise e
        
    def runPipeline(self) -> int:
        # This method is meant for local development only. Not to be used in production.
        # The Neum AI framework provides parallelization constructs through yielding
        # These should be used to run pipelines at scale.

        total_vectors_stored = 0
        for source in self.sources:
            for cloudFile in source.list_files_full():
                for localFile in source.download_files(cloudFile=cloudFile):
                    for document in source.load_data(file=localFile):
                        for chunks in source.chunk_data(document=document):
                            embeddings, embeddings_info = self.embed.embed(documents=chunks)
                            vectors_to_store = [NeumVector(id=str(uuid.uuid4()), vector=embeddings[i], metadata=chunks[i].metadata) for i in range(0,len(embeddings))]
                            total_vectors_stored += self.sink.store(vectors_to_store=vectors_to_store, pipeline_id=self.id)

        return total_vectors_stored
    
    def searchPipeline(self, query:str, number_of_results:int) -> List[NeumSearchResult]:
        vector_for_query = self.embed.embed_query(query=query)
        matches =  self.sink.search(vector=vector_for_query, number_of_results=number_of_results, pipeline_id=self.id)
        return matches

    def toPipelineModel(self):
        content_to_return = {}
        content_to_return['id'] = self.id
        content_to_return['name'] = self.name
        content = []
        for source in self.sources:
            content.append(source.to_model())
        content_to_return['sources'] = content
        content_to_return['embed'] = self.embed.to_model()
        content_to_return['sink'] = self.sink.to_model()
        content_to_return['created'] = self.created
        content_to_return['updated'] = self.updated
        if self.trigger_schedule == None:
            content_to_return['trigger_schedule'] = None
        else:
            content_to_return['trigger_schedule'] = self.trigger_schedule.to_model()

        content_to_return['latest_run'] = self.latest_run.toJson()
        content_to_return['available_metadata'] = self.available_metadata()

        return content_to_return
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['id'] = self.id
        json_to_return['name'] = self.name
        json_source = []
        for source in self.sources:
            json_source.append(source.toJson())
        json_to_return['sources'] = json_source
        json_to_return['embed'] = self.embed.toJson()
        json_to_return['sink'] = self.sink.toJson()
        json_to_return['owner'] = self.owner
        json_to_return['created'] = self.created
        json_to_return['updated'] = self.updated
        json_to_return['is_deleted'] = self.is_deleted

        if self.trigger_schedule == None:
            json_to_return['trigger_schedule'] = None
        else:
            json_to_return['trigger_schedule'] = self.trigger_schedule.toJson()

        json_to_return['latest_run'] = self.latest_run.toJson()
        json_to_return['available_metadata'] = self.available_metadata()
        return json_to_return

    def as_request(self):
        json_body = {}
        json_source = []
        for source in self.sources:
            json_source.append(source.toJson())
        json_body['sources'] = json_source
        json_body['embed'] = self.embed.toJson()
        json_body['sink'] = self.sink.toJson()
        return json_body

    def set_id(self, id: str):
        self.id = id
    
    def set_created(self, created: float):
        self.created = created

    def set_updated(self, updated: float):
        self.updated = updated

    def set_latest_run(self, pipeline_run: PipelineRun):
        self.latest_run = pipeline_run

    def set_owner(self, owner: str):
        self.owner = owner
    
    def available_metadata(self) -> List[str]:
        available_metadata = []
        for source in self.sources:
            available_metadata += source.customMetadata.keys()
            available_metadata += source.connector.selector.to_metadata
            available_metadata += source.loader.selector.to_metadata
        return available_metadata

    def as_pipeline(dct:dict):
        if dct == None:
            return None
        
        sources = dct.get("sources")
        source_value = []
        for s in sources:
            source_value.append(SourceConnector.as_source_connector(s))
        return Pipeline(
            name=dct.get("name", None),
            id=dct.get("id", None),
            sources=source_value,
            embed = as_embed(dct.get("embed")),
            sink = as_sink(dct.get("sink")),
            trigger_schedule=TriggerSchedule.as_trigger_schedule(dct.get("trigger_schedule", None)),
            latest_run=PipelineRun.as_pipeline_run(dct.get("latest_run", None)),
            created=dct.get("created", None),
            updated=dct.get("updated", None),
            owner=dct.get("owner", None),
            is_deleted=dct.get("is_deleted", False)
        )


from typing import List
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from neumai.Shared.NeumVector  import NeumVector
from neumai.Shared.NeumSearch import NeumSearchResult

class SupabaseSink(SinkConnector):
    """ Supabase Sink\n
    sink_information requires : [ database_connection ]"""

    @property
    def sink_name(self) -> str:
        return 'SupabaseSink'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ['database_connection']

    @property
    def optionalProperties(self) -> List[str]:
        return ['collection_name']

    def validation(self) -> bool:
        """Validate connector setup"""
        import vecs
        try:
            database_connection = self.sink_information['database_connection']
        except:
            raise ValueError("Required properties not set")
        try:
            vx = vecs.create_client(database_connection)
        except Exception as e:
            raise ValueError(f"Supabase connection couldn't be initialized. See exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        import vecs
        from vecs import Collection
        database_connection = self.sink_information['database_connection']
        vx = vecs.create_client(database_connection)
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")

        dimensions = len(vectors_to_store[0].vector)
        db = vx.get_or_create_collection(name=collection_name, dimension=dimensions)

        # supabase is doing some chunking here to automatically partition and batch the data, we should understand this further.
        # Also, this won't scale if there are too many items in the array
        toUpsert = []
        for i in range(0, len(vectors_to_store)):
            toUpsert.append((vectors_to_store[i].id, vectors_to_store[i].vector, vectors_to_store[i].metadata))

        db.upsert(records=toUpsert)

        # do we need to re index every time? this might be super costly. Can we check if there's an index and if not index otherwise don't?
        # db.create_index()
        # need to figure out how to do this only once. 
        vx.disconnect()
        return len(vectors_to_store)
    
    def search(self, vector: List[float], number_of_results:int, pipeline_id:str) -> List:
        import vecs
        DB_CONNECTION = self.sink_information['database_connection']
        vx = vecs.create_client(DB_CONNECTION)
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
        try:
            db = vx.get_collection(name=collection_name)
        except:
            raise Exception(f"Collection {collection_name} does not exist")
        
        try:
            results = db.query(
                data=vector,           # required
                include_metadata=True,
                include_value=True,
                limit=number_of_results,  # number of records to return
            )
        except Exception as e:
            raise Exception(f"Error querying vectors from Supabase. Exception: {e}")
        
        matches = []
        for result in results:
            matches.append(NeumSearchResult(
                id= str(result[0]),
                metadata=result[2],
                score=result[1]
            ))
        
        vx.disconnect()
        return matches
    
    def info(self, pipeline_id: str) -> NeumSinkInfo:
        import vecs
        DB_CONNECTION = self.sink_information['database_connection']
        vx = vecs.create_client(DB_CONNECTION)
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
        try:
            db = vx.get_collection(name=collection_name)
        except:
            raise Exception(f"Collection {collection_name} does not exist")
        
        number_of_vectors = db.table.select('count(*)')[0].count

        vx.disconnect()
        return NeumSinkInfo(number_vectors_stored=number_of_vectors)


from .Pipeline import (
    Pipeline
)
from .PipelineShared import (
    PipelineModel,
    PipelinesModel,
)
from .PipelineRun import (
    PipelineRun, 
    PipelineRunModel, 
    PipelineRunStatus, 
    PipelineRunTaskDetails
)
from .TriggerSchedule import TriggerSchedule
from .TriggerSyncTypeEnum import TriggerSyncTypeEnum


from neumai.DataConnectors.DataConnector import DataConnector
from neumai.DataConnectors.ConnectorHelper import as_connector
from neumai.Chunkers.Chunker import Chunker
from neumai.Chunkers.RecursiveChunker import RecursiveChunker
from neumai.Chunkers.ChunkerHelper import as_chunker
from neumai.Loaders.Loader import Loader
from neumai.Loaders.AutoLoader import AutoLoader
from neumai.Loaders.LoaderHelper import as_loader
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from typing import List, Generator
from starlette.exceptions import HTTPException
from datetime import datetime
from abc import ABC

class SourceConnector(ABC):
    def __init__(self, connector:DataConnector, chunker:Chunker = RecursiveChunker(), loader:Loader = AutoLoader(), customMetadata:dict = {}):
        self.connector = connector
        self.chunker = chunker
        self.loader = loader
        self.customMetadata = customMetadata
    
    # Add validation step to check compatibility between connectors, chunkers, loaders

    def list_files_full(self) -> Generator[CloudFile, None, None]:
        yield from self.connector.connect_and_list_full()

    def list_files_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        yield from self.connector.connect_and_list_delta(last_run=last_run)

    def download_files(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        yield from self.connector.connect_and_download(cloudFile=cloudFile)

    def load_data(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        yield from self.loader.load(file=file)

    def chunk_data(self, document:NeumDocument) -> Generator[List[NeumDocument], None, None]:
        #At this point unify all the metadata for each document to make sure it is ready downstream.
        for chunk_set in self.chunker.chunk(documents=[document]):
            chunk_set_with_custom_metadata = [NeumDocument(id=chunk.id, content=chunk.content, metadata={**chunk.metadata, **self.customMetadata, **{"text":chunk.content}}) for chunk in chunk_set]
            yield chunk_set_with_custom_metadata
    
    def validation(self) -> bool:
        core_validation = self.connector.validate() and self.loader.validate() and self.chunker.validate()
        loader_validation = self.loader.loader_name in self.connector.compatible_loaders
        return core_validation and loader_validation

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['customMetadata'] = self.customMetadata
        json_to_return['connector'] = self.connector.toJson()
        json_to_return['chunker'] = self.chunker.toJson()
        json_to_return['loader'] = self.loader.toJson()
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['customMetadata'] = self.customMetadata
        json_to_return['connector'] = self.connector.to_model()
        json_to_return['chunker'] = self.chunker.to_model()
        json_to_return['loader'] = self.loader.to_model()
        return json_to_return

    def as_source_connector(dct:dict):
        if dct == None:
            raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
        return SourceConnector(
            customMetadata=dct.get('customMetadata', {}),
            connector=as_connector(dct.get("connector", None)),
            chunker=as_chunker(dct.get("chunker", None)),
            loader=as_loader(dct.get("loader", None))
        )



from abc import ABC
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException
from typing import List

class NeumSearchResult(ABC):
    def __init__(self, id:str, metadata:dict, score:float) -> None:
        self.id =  id
        self.metadata = metadata
        self.score = score

    def as_search_result(dct:dict):
        if dct == None:
            raise HTTPException(status_code=500, detail="Value dct must be a dictionary")
        
        return NeumSearchResult(
            id=dct.get("id", None),
            metadata=dct.get("metadata", None),
            score=dct.get("score", None),
        )
    
    def toJson(self) -> dict:
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['id'] = self.id
        json_to_return['metadata'] = self.metadata
        json_to_return['score'] = self.score
        return json_to_return

class NeumSearchResultModel(JSONResponse):
    def __init__(
        self,
        content: List[NeumSearchResult],
        status_code: int = 200,
    ) -> None:
        super().__init__(content, status_code, headers=None, media_type="application/json", background=None)


from abc import abstractmethod, ABC
from typing import List, Generator
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.Selector import Selector

class Loader(ABC):
    def __init__(self, loader_information:dict = {}, selector:Selector = Selector(to_embed=[], to_metadata=[])) -> None:
        self.loader_information = loader_information
        self.selector = selector
    
    @property
    @abstractmethod
    def loader_name(self) -> str:
        pass
    
    @property
    @abstractmethod
    def requiredProperties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optionalProperties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def availableMetadata(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def availableContent(self) -> List[str]:
        pass

    @abstractmethod
    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""

    @abstractmethod
    def validate(self) -> bool:
        """Validate if the loader is correctly configured"""

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['loader_name'] = self.loader_name
        json_to_return['loader_information'] = self.loader_information
        json_to_return['selector'] = self.selector.toJson()
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different than toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['loader_name'] = self.loader_name
        json_to_return['loader_information'] = self.loader_information
        json_to_return['selector'] = self.selector.to_model()
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['requiredProperties'] = self.requiredProperties
        json_to_return['optionalProperties'] = self.optionalProperties
        json_to_return['availableMetadata'] = self.availableMetadata
        json_to_return['availableContent'] = self.availableContent
        return json_to_return


from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from neumai.Shared.NeumVector  import NeumVector
from neumai.Shared.NeumSearch import NeumSearchResult
from neumai.SinkConnectors.SinkConnector import SinkConnector
from typing import List

class QdrantSink(SinkConnector):
    """ Qdrant Sink\n
    sink_information requires : [ 'url', 'api_key' ]"""

    @property
    def sink_name(self) -> str:
        return 'QdrantSink'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ['url', 'api_key']

    @property
    def optionalProperties(self) -> List[str]:
        return ['collection_name']

    def validation(self) -> bool:
        """Validate connector setup"""
        from qdrant_client import QdrantClient
        try:
            url = self.sink_information["url"]
            api_key = self.sink_information["api_key"]
        except:
            raise ValueError("Required properties not set")
        try:
            qdrant_client = QdrantClient(
                url=url, 
                api_key=api_key,
            )
        except Exception as e:
            raise ValueError(f"Qdrant connection couldn't be initialized. See exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        from qdrant_client.http.models import Distance, VectorParams
        from qdrant_client.http.models import PointStruct
        from qdrant_client.http.models import UpdateStatus
        from qdrant_client import QdrantClient

        url = self.sink_information["url"]
        api_key = self.sink_information["api_key"]
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
        qdrant_client = QdrantClient(
            url=url, 
            api_key=api_key,
        )
        qdrant_client.recreate_collection(
            collection_name=collection_name,
            vectors_config=VectorParams(size=1536, distance=Distance.DOT)
        )
        points = [PointStruct(id=vector.id, vector=vector.vector, payload=vector.metadata) for vector in vectors_to_store]
        operation_info = qdrant_client.upsert(
            collection_name=collection_name,
            wait=True,
            points=points
        )
        if(operation_info.status == UpdateStatus.COMPLETED):
            return  len(points)
        raise Exception("Qdrant storing failed. Try again later.")
    
    def search(self, vector: List[float], number_of_results: int, pipeline_id: str) -> List:
        from qdrant_client import QdrantClient
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        collection_name = self.sink_information['collection_name']
        try:
            qdrant_client = QdrantClient(
                url=url, 
                api_key=api_key,
            )
            search_result = qdrant_client.search(
                collection_name=collection_name,
                query_vector=vector, 
                with_payload= True,
                limit=number_of_results
            )
        except Exception as e:
            raise(f"Failed to query Qdrant. Exception - {e}")
        
        matches = []
        for result in search_result:
            matches.append(
                NeumSearchResult(
                    id=result.id,
                    metadata=result.payload,
                    score=result.score
                )
            )
        return matches
    
    def info(self, pipeline_id: str) -> NeumSinkInfo:
        from qdrant_client import QdrantClient
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        collection_name = self.sink_information['collection_name']
        try:
            qdrant_client = QdrantClient(
                url=url, 
                api_key=api_key,
            )
            collection_info = qdrant_client.get_collection(collection_name=collection_name)
            return(NeumSinkInfo(number_vectors_stored=collection_info.indexed_vectors_count))
        except Exception as e:
            raise(f"Failed to get information from Qdrant. Exception - {e}")